#!/usr/bin/env python
# coding: utf-8

"""

    pyPhelix.py
    
    Phelix is a highly efficient algorithm that, in a single 
    pass of the data, both encrypts the data and generates a 
    MAC.  Phelix is a new algorithm submitted to eSTREAM by 
    Doug Whiting, Bruce Schneier, Stephan Lucks and Frdric 
    Muller.  Phelix is new and not yet field proven.  For now, 
    the use of Phelix should be limited to selected low value 
    assets where throughput is especially critical.  Phelix is 
    in the Public Domain.  See: 
      http://www.ecrypt.eu.org/stream/phelixp2.html
    
    Compile phelix.c and install as a shared library, not a
    Python extension.  See README.txt file for more information.
    
    pyPhelix.py is a simple ctypes Python wrapper written by 
    Larry Bugbee.  It is experimental software and intended 
    for educational purposes only.  THIS SOURCE IS PROVIDED 
    WITHOUT WARRANTY OR GUARANTEE OF ANY KIND.  USE AT YOUR 
    OWN RISK.  To make experimentation less cumbersome, 
    pyPhelix.py is freeware.    
    
    Some Phelix parameter values, as I understand them...
        design strength     128 bits
        key length          0-256 bits in 8-bit increments
        IV, aka nonce       128 bits, always
        MAC size            0-128 bits in 1-bit increments
        chunk size          any size in 32-bit increments
    
    Usage:
        ph = Phelix(key, nonce, macSize)
        ciphertext = ph.encryptBytes(plaintext)
        mac = ph.getMAC()
    
    On my 1.5 GHz PPC G4 miniMac with 1 GB RAM, encrypting and 
    writing a 10 MB file took less than a second.  A 118 MB 
    file took just over 7 seconds.  Excluding I/O and other 
    setup, Phelix is encrypting and hashing about at 36 MB per 
    second.  Decryption is a little slower at about 31 MB per 
    second.  All this is with the "unoptimized" C reference 
    version of Phelix.  By contrast, using libTomCrypt I get...
        Twofish   21.3 MB per sec
        AES       19.2 MB per sec
        blowfish  19.1 MB per sec
        cast5     17.6 MB per sec
        anubis    15.1 MB per sec
        xtea      13.8 MB per sec
        3des       7.6 MB per sec
        
    Enjoy,
      
    Larry Bugbee
    bugbee@seanet.com
    March 12, 2007
    
"""

from ctypes import *
import sys

_version = '0.07'

#---------------------------------------------------------------

class Phelix(object):
    def __init__(self, key='', IV='', macSize=128):
        self._IVlen   = 128         # must be 128 bits
        self._macsize = macSize     # 0-128 bits (8-bit incr)
        
        self._ctx = c_buffer(132)   # libPhelix.PhelixContextSize()
        if not key:
            key = ''                # be sure it is type string
        self.setKey(key)
        if IV:
            self.setIV(IV)

    def setKey(self, key):
        libPhelix.PhelixSetupKey(byref(self._ctx), 
                                 key, 
                                 len(key)*8, 
                                 self._IVlen,
                                 self._macsize)
        
    def setIV(self, nonce):
        libPhelix.PhelixSetupNonce(byref(self._ctx), 
                                   nonce[:self._IVlen/8])
    setnonce = setIV

    def encryptBytes(self, data):
        ciphertext = c_buffer(len(data))
        libPhelix.PhelixEncryptBytes(byref(self._ctx), 
                                     data, 
                                     byref(ciphertext), 
                                     len(data))
        return ciphertext.raw

    def decryptBytes(self, data):
        plaintext = c_buffer(len(data))
        libPhelix.PhelixDecryptBytes(byref(self._ctx), 
                                     data, 
                                     byref(plaintext), 
                                     len(data))
        return plaintext.raw
        
    # AAD: 'Additional Authentication Data'
    # to include additional data in the MAC value without 
    # encrypting it (example: a packet header)
    def processAAD(self, data):
        libPhelix.PhelixProcessAAD(byref(self._ctx), 
                                   data, 
                                   len(data))
        
    def getMAC(self, mac=None):
        if self._macsize%8 == 0:
            cb_mac = c_buffer(self._macsize/8)
        else:
            cb_mac = c_buffer(self._macsize/8 + 1)
        libPhelix.PhelixFinalize(byref(self._ctx), 
                                 byref(cb_mac))
        return cb_mac.raw
    finalize = getMAC

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# functions

def processPacket(action, key, nonce, src, macSize, aad):
    ph = Phelix(key, macSize=macSize)
    cb_dst = c_buffer(len(src))
    if ph._macsize%8 == 0:
        cb_mac = c_buffer(ph._macsize/8)
    else:
        cb_mac = c_buffer(ph._macsize/8 + 1)
    libPhelix.PhelixProcessPacket(action,
                                  byref(ph._ctx), 
                                  nonce, 
                                  aad, 
                                  len(aad), 
                                  src, 
                                  byref(cb_dst), 
                                  len(src), 
                                  cb_mac)
    return cb_dst.raw, cb_mac.raw

def encryptPacket(key, nonce, src, macSize=128, aad=''):
    return processPacket(0, key, nonce, src, macSize, aad)

def decryptPacket(key, nonce, src, macSize=128, aad=''):
    return processPacket(1, key, nonce, src, macSize, aad)

def getCompilerName():
    libPhelix.PhelixCompiler_Name.restype = c_char_p
    return libPhelix.PhelixCompiler_Name()
    
def getIncremental_CodeSize():
    return libPhelix.PhelixIncremental_CodeSize()
    
def getProcessPacket_CodeSize():
    return libPhelix.PhelixProcessPacket_CodeSize()


#---------------------------------------------------------------
# load the shared library into ctypes

def loadLib(name):
    # prefix might need to change for some platforms       ???
    prefix = ''
    # get the correct library suffix
    libsuffixes = {'darwin': '.so',     # .dylib           ???
                   'linux':  '.so', 
                   'linux2': '.so', 
                   'win32':  '.dll'}    # .so  .lib        ???
    try:
        libsuffix = libsuffixes[sys.platform]
    except:
        raise Exception('library suffix for "%s" is what?' % 
                                                  sys.platform)
    libname = prefix+'lib'+name+libsuffix
    return CDLL(libname)                # load the library


libPhelix = loadLib('phelix')
libPhelix.PhelixInit()          # one-time init


#---------------------------------------------------------------
#---------------------------------------------------------------
# a quick self-test.  Invoke with:  python pyPhelix.py

def test():

    message = 'Kilroy was here!  ...and there.'
    key     = 'myKey'               # do better in real life
    nonce   = 'aNonce'              # do better in real life
    IV      = (nonce+'*'*16)[:16]   # force to exactly 128 bits
    header  = '<< this is a header >>'
    print '       pyPhelix version: %s' % _version
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # simple encrypt and get MAC
    print '  simple encypt and MAC:',

    ph = Phelix(key, IV)
    ciphertext = ph.encryptBytes(message)   # do whole message
    mac1 = ph.getMAC()
    
    # decrypt and get MAC
    ph2 = Phelix(key, IV)
    # if multiple chunks, chunk in increments of 4 bytes
    plaintext  = ph2.decryptBytes(ciphertext[:12]) # do 1st 12
    plaintext += ph2.decryptBytes(ciphertext[12:]) # do remainder
    mac2 = ph2.getMAC()
    
    if plaintext == message and mac2 == mac1:
        print '  *** good ***'
    else:
        print '  *** bad ***'
    
    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # add a header and retest
    print '  add header and encypt:',

    # encrypt and get MAC
    ph3 = Phelix(key, IV)
    ph3.processAAD(header)
    ciphertext = ph3.encryptBytes(message)   # do whole message
    mac3 = ph3.getMAC()
    
    # decrypt and get MAC
    ph4 = Phelix(key, IV)
    ph4.processAAD(header)
    # if multiple chunks, chunk in increments of 4 bytes
    plaintext  = ph4.decryptBytes(ciphertext[:12]) # do 1st 12
    plaintext += ph4.decryptBytes(ciphertext[12:]) # do remainder
    mac4 = ph4.getMAC()

    if plaintext == message and mac4 == mac3:
        print '  *** good ***'
    else:
        print '  *** bad ***'

    
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # test the packet functions
    print '       packet functions:',
    macSize = 128
    
    # encrypt and get MAC
    ciphertext, mac5 = encryptPacket(key, IV, message, macSize, header)
    
    # decrypt and get MAC
    plaintext, mac6 = decryptPacket(key, IV, ciphertext, macSize, header)

    if plaintext == message and mac5 == mac3 and mac6 == mac5:
        print '  *** good ***'
    else:
        print '  *** bad ***'

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

if __name__ == '__main__':
    test()

#---------------------------------------------------------------
#---------------------------------------------------------------
#---------------------------------------------------------------
